/*
Copyright (C) 2009 -  Claudio Salvadori and Paolo Pagano
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
/**
 *\file DAQ.cpp
 *\date  Created on: 30-06-2009
 *\author  Author: Claudio Salvadori and Paolo Pagano
 *\brief  DAQ main functions */

#include <vector>
#include <string>
#include "DAQ_struct.h"
#include "DAQ_enum.h"
#include "DAQError.h"
#include <signal.h>
#include "CDaq.h"

#include "Show_image.h"
#include "stdio.h"

#include "globals.h"

#ifdef USE_ARG_PARSING
#include <argp.h>
#endif /* USE_ARG_PARSING */

using namespace std;

static char* gXmlFile = 0;
#define PROG_TAG "\nMy Serial Debug :  "

void close_handler(int signo);
string error_handler(daq_ret error);


#ifdef USE_ARG_PARSING
const char *argp_program_version = "DAQ 1.0";
const char *argp_program_bug_address = "<p.pagano@sssup.it>";

/* Program documentation. */
const char *doc = "DAQ -- Data AcQuisition System\
\vTODO write a detailed description\nStart data acquisition using the CONFIG xml file.";

/* A description of the arguments we accept. */
const char *args_doc = "CONFIG";

/* The options we understand. */
static struct argp_option options[] = {
	{"verbose",  'v', 0,      0,  "Produce verbose output" },
	//{"quiet",    'q', 0,      0,  "Don't produce any output" },
	//{"silent",   's', 0,      OPTION_ALIAS },
	{"numevent",   'n', "NEVT", OPTION_ARG_OPTIONAL,
	"Output to FILE instead of standard output" },
	{"rootfile",   'f', "FILE", OPTION_ARG_OPTIONAL,
	"Output to FILE instead of standard output" },
	{ 0 }
};

/* Used by main to communicate with parse_opt. */
struct arguments
{
	char *xml_file_name;
	char *rootfilename;
	int silent, verbose;
	unsigned long nevents;
	//char *output_file;
};

/* Parse a single option. */
static error_t parse_opt(int key, char *arg, struct argp_state *state)
{
	/* Get the input argument from argp_parse, which we
	   know is a pointer to our arguments structure. */
	struct arguments *arguments = (struct arguments *) state->input;
	
	switch (key) {
	//case 'q': case 's':
	//	arguments->silent = 1;
	//	break;
	case 'v':
		arguments->verbose = 1;
		#define DEBUG
		break;
	case 'n':
		arguments->nevents = atoi(arg);
		break;
	case 'f':
		arguments->rootfilename = arg;
		break;
	case ARGP_KEY_ARG:
		if (state->arg_num >= 1)/* Too many arguments. */
	        	argp_usage(state);
		arguments->xml_file_name = arg;
		break;
	case ARGP_KEY_END:
		if (state->arg_num < 1) /* Not enough arguments. */
			argp_usage(state);
		break;
	default:
		return ARGP_ERR_UNKNOWN;
	}
	return 0;
}

/* Our argp parser. */
static struct argp argp = {options, parse_opt, args_doc, doc};

#endif /* USE_ARG_PARSING */
void initStorage(char *rootfilename){
	rootfile_ = new TFile(rootfilename, "RECREATE"); 
	cerr << "ROOTFile instantiated :" << rootfile_ << endl;
}

int main(int argC, char* argV[]){
	static int ev =0;
// 	static int endnum = 1000000;
// 	if (argC > 2){
// 		endnum = atoi(argV[2] );
// 	}
	
	//installing signals
	
#ifdef USE_ARG_PARSING
	struct arguments arguments;

	/* Default values. */
	arguments.xml_file_name = 0;
	arguments.nevents = 1000000; // default is infinite
	arguments.rootfilename = (char*)"test.root";//default_name;
	
	/* Parse our arguments; every option seen by parse_opt will
	   be reflected in arguments. */
	argp_parse (&argp, argC, argV, 0, 0, &arguments);
	gXmlFile = arguments.xml_file_name;
#else
	gXmlFile = argV[1]; //xml file name
#endif
	initStorage(arguments.rootfilename);
	struct sigaction nsig,vsig;
        nsig.sa_handler = close_handler;
        nsig.sa_flags=0;
        sigemptyset(&(nsig.sa_mask));

        if (sigaction(SIGINT, &nsig,&vsig)==-1) {
                cout << PROG_TAG << "Handler Install Error " << endl << endl;
                exit(EXIT_FAILURE);
	}
	
	CDaq *DAQ = NULL;
	vector< vector<SData> >* data_formatted = NULL;
	
	daq_ret help = D_RET_OK;
	
	#ifdef ROOT_SYSTEM
    	// Create a ROOT Tree and one superbranch
	TApplication theApp("App", &argC, argV);

	event = 0;
	tree = new TTree("T","An example of a ROOT tree for the DAQ");
	Int_t bufsize = 64000;
	tree->SetAutoSave(1000000);  // autosave when 1 Gbyte written
	int split = 1;
	if (split)  bufsize /= 4;
	event = new DAQEvent();
	TTree::SetBranchStyle(1);
	TBranch *branch = tree->Branch("event", &event, bufsize,1);
	branch->SetAutoDelete(kFALSE);
	if(split >= 0) tree->BranchRef();
	#endif
	
	DAQ = new CDaq(gXmlFile);
	help = DAQ->init();

	if(help != D_RET_OK){
		//The best thing is to add a function to write the errors descriptions
		cerr << "CDaq::init() returns: " << error_handler(help) << endl;
		exit(1);
	}
	printf("CDaq::init() returns: OK \n");
	//Creation of the data formatted vector
	data_formatted = new vector< vector<SData> >;
	
	static unsigned long num = 0;
	fprintf(stderr, "Reading %lu events\n", arguments.nevents);
	if(entries_ != 0)
		arguments.nevents = entries_;
		
	while(num < arguments.nevents){
		//reading and formatting data
		help = DAQ->getData(data_formatted);
		if(help != D_RET_OK){
			//The best thing is to add a function to write the errors descriptions
			cerr << "CDaq::getData(...) returns: " << error_handler(help) << endl;
			exit(1);
		}

		printf("CDaq::getData(...) returns: OK for event number %lu\n", num);

		event->Build(ev++, data_formatted);
		tree->Fill();  //fill the tree
		num ++;
		
	}
	
	return 0;
}
void close_handler(int signo){	
        cout << PROG_TAG << "Closing aa connection ... ";
        cout << "Done!";
        cout << PROG_TAG << "Exiting ... Bye! " << endl << endl;
        exit(EXIT_SUCCESS);
}

string error_handler(daq_ret error){
	switch(error){
	case D_RET_OK:	
		return "Ok";
		break;
	
	case D_INIT_ERR:
		return "Error in parser init";
		break;

	case D_PARSE_ERR:
		return "Error in parsing";
		break;

	case D_XML_PARSER_ERR:
		return "Error in parsing";
		break;

	case D_XML_P_FILE_ERR:
		return "Syntax error in file.xml";
		break;

	case D_XML_P_DOM_ERR:
		return "Error in DOM parsing";
		break;

	case D_XML_P_UN_ERR:
		return "Unexpected exception during parsing";
		break;

	case D_P_STYLE_ERR:
		return "Style error";
		break;

	case D_P_TITLE_ERR:
		return "Title error";
		break;

	case D_P_STREAM_ERR:
		return "Stream node searching error";
		break;

	case D_P_STREAM_T_ERR:
		return "Stream type error";
		break;

	case D_P_STREAM_N_ERR:
		return "Stream name error";
		break;

	case D_P_STREAM_FT_ERR:
		return "Stream file type error";
		break;

	case D_P_STREAM_C_ERR:
		return "Stream channel/port error";
		break;

	case D_P_PKT_ERR:
		return "Packet node searching error";
		break;

	case D_P_PKT_S_ERR:
		return "Packet size error";
		break;

	case D_P_DTA_ERR:
		return "Data node searching error";
		break;

	case D_P_DTA_N_ERR:
		return "Data name error";
		break;

	case D_P_DTA_P_ERR:
		return "Data position error";
		break;

	case D_P_DTA_S_ERR:
		return "Data size error";
		break;

	case D_P_DTA_T_ERR:
		return "Data type error";
		break;

	case D_GENERIC_ERR:
		return "Generic error";
		break;

	case D_VAL_MDATA_ERR:
		return "Validate metadata error";
		break;

	case D_P_S_STREAM_ERR:
		return "Serial stream node searching error";
		break;

	case D_P_BAUD_RATE_ERR:
		return "Baud rate error";
		break;

	case D_P_DATA_BITS_ERR:
		return "Data bits error";
		break;

	case D_P_PARITY_ERR:
		return "Parity bits error";
		break;

	case D_P_STOP_ERR:
		return "Stop bits error";
		break;

	case D_P_HAND_ERR:
		return "Handshaking node searching error";
		break;

	case D_P_HAND_P_ERR:
		return "Handshaking packet node searching error";
		break;

	case D_P_H_SEND_ERR:
		return "Send packet error";
		break;

	case D_P_O_STREAM_ERR:
		return "Option stream node searching error";
		break;

	case D_P_O_TYPE_ERR:
		return "Option type error";
		break;

	case D_P_O_START_ERR:
		return "Start character error";
		break;

	case D_P_O_END_ERR:
		return "End character error";
		break;

	case D_P_O_ESC_ERR:
		return "Escape character error";
		break;

	case D_P_H_ACK_ERR:
		return "Ack packet error";
		break;

	case D_P_H_NACK_ERR:
		return "Nack packet error";
		break;

	case D_P_H_STYLE_ERR:
		return "Handshaking style error";
		break;
		
	case D_P_H_LENGTH_ERR:
		return "File: header length error";
		break;
		
	case D_P_N_ENTRIES_S_ERR:
		return "File: number of entries start byte error";
		break;
		
	case D_P_N_ENTRIES_L_ERR:
		return "File: number of entries length error";
		break;
		
	case D_P_T_LENGTH_ERR:
		return "File: trailer length error";
		break;
		
	case D_OPTION_ERR:
		return "Option struct doesn't fill correctly";
		break;

	case D_D_EMPTY_ERR:
		return "Empty input buffer (decode)";
		break;

	case D_D_ESCAPE_ERR:
		return "Escape character in the wrong position";
		break;

	case D_M_EMPTY_ERR:
		return "Empty input buffer (divide in messages)";
		break;

	case D_M_LENGTH_ERR: 
		return "Wrong packet size (divide in messages)";
		break;
	
	case D_F_OPEN_ERR:
		return "Error in file opening";
		break;

	case D_F_READ_ERR:
		return "Error in file reading";
		break;

	case D_F_R_LENGTH_ERR:
		return "Error: read shortest packet in fileStream name error";
		break;

	case D_F_R_START_ERR:
		return "Error in reading file: start character doesn't find";
		break;

	case D_F_R_END_ERR:
		return "Error in reading file: end character doesn't find";
		break;

	case D_F_EOS:
		return "OK end of stream for files";
		break;
	
	case D_S_PRESET_MISS_ERR:
		return "Error: missing the serial preset";
		break;

	case D_S_OPEN_ERR:
		return "Error in serial open";
		break;

	case D_S_G_CHAR_ERR:
		return "Error: serial get char";
		break;

	case D_S_P_CHAR_ERR:
		return "<Error: serial put char";
		break;

	case D_S_READ_ERR:
		return "Error in receiving from serial";
		break;

	case D_S_WRITE_ERR:
		return "Error in sending by serial";
		break;

	case D_S_LENGHT_ERR:
		return "Error: read shortest/longest packet in file";
		break;

	case D_S_R_TIMEOUT_ERR:
		return "Error: serial receive timeout error";
		break;

	case D_S_T_TIMEOUT_ERR:
		return "Error: serial transmit timeout error";
		break;

	case D_S_NACK:
		return "Nack received during handshking (serial)";
		break;

	case D_S_GEN_ERR:
		return "Generic error(serial)";
		break;
	
	case D_P_INIT_ERR:
		return "Error in pipe init";
		break;

	case D_P_G_CHAR_ERR:
		return "Error: pipe get char";
		break;

	case D_P_P_CHAR_ERR:
		return "Error: pipe put char";
		break;

	case D_P_READ_ERR:
		return "Error in receiving from pipe";
		break;

	case D_P_WRITE_ERR:
		return "Error in sending by pipe";
		break;

	case D_P_LENGHT_ERR:
		return "Error: read shortest/longest packet in file";
		break;

	case D_P_R_TIMEOUT_ERR:
		return "Error: pipe receive timeout error";
		break;

	case D_P_T_TIMEOUT_ERR:
		return "Error: pipe transmit timeout error";
		break;

	case D_P_NACK:
		return "Nack received during handshking (pipe)";
		break;

	case D_P_GEN_ERR:
		return "Generic error(pipe)";
		break;

	case D_P_NOTIMPL:
		return "Pipe is not implemented on Linux. Use Socket!!";
		break;

	case D_SO_INIT_ERR:
		return "Error in initializing the socket";
		break;
	case D_SO_INIT_V_ERR:
		return "Error in initializing the socket(not correct version)";
		break;
	case D_SO_INIT_R_ERR:
		return "Error in initializing the socket(resolving server)";
		break;
	case D_SO_INIT_C_ERR:
		return "Error in initializing the socket(create socket error)";
		break;
	case D_SO_INIT_CONN_ERR:
		return "Error in initializing the socket(client connection error)";
		break;
	case D_SO_WRITE_ERR:
		return "Socket write error";
		break;
	case D_SO_LENGHT_ERR:
		return "Error: read shortest/longest packet from socket";
		break;
	case D_SO_NACK:
		return "Socket: Nack received during handshking";
		break;
	case D_SO_GEN_ERR:
		return "Socket: Generic error";
		break;
	case D_SO_READ_ERR:
		return "Socket read error";
		break;
	case D_SO_CLOSED_ERR:
		return "Socket connection closed";
		break;

	/**********************************************************************************************/
	/*       ADD HERE THE RETURN ERRORS MESSAGE FROM THE NEW COMMUNICATION DEVICE CLASS           */
	/**********************************************************************************************/

	case D_META_ERR:
		return "Metadata vector is empty";
		break;

	case D_H_NULL_ERR:
		return "Header vector is pointed by a NULL pointer";
		break;

	case D_H_EMPTY_ERR:
		return "Header vector is empty";
		break;

	case D_D_NULL_ERR:
		return "Data vector is pointed by a NULL pointer";
		break;

	case D_F_EMPTY_ERR:
		return "Data vector is empty";
		break;
		
	case D_P_ENDIANESS_ERR:
		return "Endianess error";
		break;
	}
	return "OK";
}

