/*
Copyright (C) 2009 -  Claudio Salvadori
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/**
 *\file CXmlParser.cpp
 *\brief  XML parsing and analysis development functions
 */

#include "CXmlParser.h"
#include <stdio.h>

////////////////////////////////////////////////////////////////////////////////////////////////////////
// 									CXMLParsefmethods					        //
///////////////////////////////////////////////////////////////////////////////////////////////////////

//Constructor
CXMLParser::CXMLParser()
{
	mParser=NULL;
	mErrorHandler=NULL;
	g_Transcoder=NULL;
	g_bTranscoder=false;
}

CXMLParser::~CXMLParser()
{
	delete mParser;
	delete mErrorHandler;
	delete g_Transcoder;
}

//Parser init method
xml_parser_ret_value CXMLParser::Init(void)
{
	try{
		XMLPlatformUtils::Initialize();
    	}

	catch(const XMLException &toCatch){
		#ifdef DEBUG
		char disp[256];
		char *msg=strX(toCatch.getMessage());

		sprintf(disp,"Error during Xerces-c Initialization\n%s", msg);
		delete [] msg;
		#endif
	
		return R_XML_INIT_ERR;
	}
	 
	if ( (mParser = new XercesDOMParser) == NULL){
		#ifdef DEBUG
		char disp[256];
		sprintf(disp, "Error: Can not create parser object\n");
		#endif
		return R_XML_INIT_ERR;
	}
		
 	mParser->setValidationScheme (XercesDOMParser::Val_Always);
	//mParser->setDoValidation(true);
	mParser->setDoNamespaces(true);
	mParser->setDoSchema(true);
	mParser->setExternalNoNamespaceSchemaLocation("input.xsd"); 



	mErrorHandler = new DOMTreeErrorReporter();
	mParser->setErrorHandler(mErrorHandler);    


	mParser->setCreateEntityReferenceNodes(false);

	return R_XML_OK;
}

//Parse method
xml_parser_ret_value CXMLParser::Parse(char *fn, DOMNode* &doc){
    
    	try{
		mParser->parse(fn);
		int errorCount = mParser->getErrorCount();
		if (errorCount > 0){
		    #ifdef DEBUG
		    printf("Error during parsing: error count = %d \n", errorCount);
		    #endif
		    return R_XML_P_FILE_ERR;
		}
    	}

   	catch (const XMLException& e){
		printf("11\n");
		#ifdef DEBUG
		char disp[256];
		char *msg=strX(e.getMessage());
		sprintf(disp,"Error during parsing %s\n%s",fn,msg);
		printf("%s\n", disp);
		delete [] msg;
		#endif 

		return R_XML_PARSER_ERR;
    	}

    	catch (const DOMException& e){
		#ifdef DEBUG
		char disp[256];
		sprintf(disp,"DOM Error during parsing %s\nDOMException code: %d",fn,e.code);
		printf("%s", disp);
		#endif 
		return R_XML_P_DOM_ERR;
    	}
    	catch (...){
		#ifdef DEBUG
		char disp[256];
		sprintf(disp,"Unexpected exception during parsing %s",fn);
		printf("%s", disp);
		#endif 

		return R_XML_P_UN_ERR;
    	}

	doc=mParser->getDocument();

	return R_XML_OK;
}

// Check if the actual node corresponds with the required one
bool CXMLParser::IsTargetNode(DOMNode* node, char * targetname)
{
	bool ret=false;
	int nodetype = node->getNodeType();
	char *strname=XMLStr2CStr(node->getNodeName());

	if(strcmp(strname,targetname)==0 && nodetype==DOMNode::ELEMENT_NODE)
		ret=true;

	delete [] strname;
	return ret;
}

/* Get the child node of curRoot*/
bool CXMLParser::GetChildNode(DOMNode* curRoot, char * targetname, DOMNode* &targetnode)
{
    	for(DOMNode* child = curRoot->getFirstChild(); child != NULL; child = child->getNextSibling()){
		if(IsTargetNode(child,targetname)){
			targetnode=child;
			return true;
		}
	}

	return false;
}

//Search a target node inside the tree
bool CXMLParser::FindNodeRootat(DOMNode* curRoot, char * targetname, DOMNode* &targetnode)
{
	DOMNode* child;
	if(IsTargetNode(curRoot,targetname)){
		targetnode=curRoot;
		return true;
	}

    	for(child = curRoot->getFirstChild(); child != NULL; child = child->getNextSibling()){
		if(IsTargetNode(child,targetname)){
			targetnode=child;
			return true;
		}
	}

	//recurse
   	 for(child = curRoot->getFirstChild(); child != NULL; child = child->getNextSibling()){
		if(FindNodeRootat(child,targetname,targetnode))
			return true;
	}

	return false;
}

//Get the value of the attribute called "attrname" of the current node (DOMNode* node)
bool CXMLParser::GetAttribute(DOMNode* node, char * attrname, char* attrvalue)
{
	bool ret=false;
	char *name, *value;

	#ifdef DEBUG
	printf("Node type: %d\n", node->getNodeType());
	printf("Searching for attribute: %s\n", attrname);
	#endif

	DOMNamedNodeMap* attributes;
	attributes = node->getAttributes();
	if(attributes != NULL )
	{	
		int len = attributes->getLength() ;
		for ( int i = 0; i < len; i++ ){
			DOMNode*  attribute = attributes->item(i);
			name=XMLStr2CStr(attribute->getNodeName());
			value=XMLStr2CStr(attribute->getNodeValue());

			if(strcmp(name,attrname)==0){
				strcpy(attrvalue,value);
				ret=true;
			}
			delete [] name;
			delete [] value;
			if(ret)
				return true;
		}
	}
	return false;
}

//Find the sibling node, type == 1 (DOMElement)
DOMNode* CXMLParser::Get_NextSibling(DOMNode *node){
	DOMNode *x;
	x = node->getNextSibling();
	
	if(x == NULL)
		return NULL;

	while (x->getNodeType() != 1){
 		x = x->getNextSibling();
		if(x == NULL)
			return NULL;
	}
	return x;
}

//Get the text of the current node (DOMNode* node)
bool CXMLParser::GetText(DOMNode* node,char* text)
{
	char *value;

    	for(DOMNode* child = node->getFirstChild(); child != NULL; child = child->getNextSibling())
	{
		if(child->getNodeType()==DOMNode::TEXT_NODE){
			value=XMLStr2CStr(child->getNodeValue());
			strcpy(text,value);
			delete [] value;
			return true;
		}
	}
	return false;
}

//Change string encoder
void CXMLParser::setTranscoder(char *codepage)
{
	XMLTransService::Codes resCode;
	g_Transcoder=XMLPlatformUtils::fgTransService->makeNewTranscoderFor(codepage, resCode, 65536);
	g_bTranscoder=true;
}

//Method to decode string from XMLCh* to char*
char* CXMLParser::XMLStr2CStr(const XMLCh* s)
{
	if(!g_bTranscoder)
		return strX(s);
	else{
		unsigned int tmp;
		unsigned int len=XMLString::stringLen(s);

		char *ret=new char[2*len+4];
		if(ret==NULL)
			return NULL;
		memset(ret,0,2*len+4);
		g_Transcoder->transcodeTo((const XMLCh*)s, (XMLSize_t)len, (XMLByte*)ret, (XMLSize_t)2 * len,(XMLSize_t&)tmp,XMLTranscoder::UnRep_Throw);
		return ret;
	}
}


